
module Phase4 where
import Prelude          as P
import Dictionnaire     as D
import Data.Map.Strict  as M
import Data.Set         as S
import Control.Monad
import Data.Hashable
import Autorise


type Hash = Int
type Salt = Int
type Nom  = Int

type Transitions = Dictionnaire Char Nom
data Etat = Etat { getFinal :: Bool
                 , getTrans :: Transitions
                 }
            deriving Show
data AutomateFini = AutomateFini { getEtats   :: (Map Nom Etat)
                                 , getInitial :: Nom
                                 }
                    deriving Show



-- Reconnaissance d'un mot par un automate
reconnise :: AutomateFini -> String -> Bool
reconnise (AutomateFini etats nom) s = reconniseE (etats ! nom) s  
     where
       reconniseE e ""         = getFinal e
       reconniseE e (c:w)      = reconniseM (rechercherParCle (getTrans e) c) w
       reconniseM Nothing w    = False
       reconniseM (Just nom) w = reconniseE (etats ! nom) w

-- créer un automate a partir d'une liste d'états avec leur booleen d'acceptation et d'une liste de transitions
-- l'état initial est le premier nom de la liste de noms
newAutomaton :: [(Nom,Bool)] -> [(Nom,Char,Nom)] -> AutomateFini
newAutomaton noms tr = AutomateFini (creerEtats noms tr) (fst (head noms))

creerEtats :: [(Nom,Bool)] -> [(Nom,Char,Nom)] -> Map Nom Etat
creerEtats noms tr = mapWithKey (creerEtat tr) (M.fromList noms)

creerEtat :: [(Nom,Char,Nom)] -> Nom -> Bool -> Etat
creerEtat tr e b = Etat b (insererListe creer transitions)
   where
     tr_e = P.filter (\(e',_,_)->e==e') tr
     transitions = fmap (\(_,c,n)->(c,n)) tr_e

-- type des automate infinis du TP2
data EtatD = EtatD Bool (Char->EtatD)
type AutomateD = EtatD

-- projeter sur le type du TP2
forgetFiniteness :: AutomateFini -> AutomateD
forgetFiniteness (AutomateFini etats init) = forgetFinitenessEtat (etats ! init)
   where
   forgetFinitenessEtat (Etat b tr) =
       EtatD b (forgetFinitenessTransition tr)
   forgetFinitenessTransition tr c =
       let nom = recherche tr c
       in let etat = etats ! nom
       in forgetFinitenessEtat etat


-- projeter sur le type du TP2
forgetFinitenessSafe :: AutomateFini -> AutomateD
forgetFinitenessSafe (AutomateFini etats init) = forgetFinitenessEtat (etats ! init)
   where
   forgetFinitenessEtat (Etat b tr) =
       EtatD b (enleverMaybe.(forgetFinitenessTransition tr))
   forgetFinitenessTransition tr c = do
       nom  <- rechercherParCle tr c
       etat <- etats !? nom
       return (forgetFinitenessEtat etat)
   enleverMaybe (Just e) = e
   enleverMaybe Nothing = EtatD False (\c-> enleverMaybe Nothing)



-- version non-déterministe
type NomND = Int
type TransND = Dictionnaire Char [NomND]
data EtatND = EtatND { getFinalND :: Bool
                     , getTransND :: TransND
                     }
                     deriving Show
data AutomateFiniND = AutomateFiniND { getEtatsND :: Map NomND EtatND
                                     , getInitND   :: [NomND]
                                     }
                     deriving Show

-- fonction de determinisation
determinise :: AutomateFiniND -> AutomateFini
determinise (AutomateFiniND etatsND initND) = AutomateFini etats init
   where
   init  = nommer initND
   etats = determiniseAux etatsND M.empty [initND] 

determiniseAux :: (Map NomND EtatND) -> Map Nom Etat -> [[NomND]] -> Map Nom Etat
determiniseAux etatsND acc [] = acc
determiniseAux etatsND acc (noms:l)
        | M.member nomDet acc = determiniseAux etatsND acc l
	| otherwise           = determiniseAux etatsND (M.insert nomDet etat acc) ((D.toListVal ft)++l)
	where
	  nomDet   = nommer noms
	  etat     = Etat (estTerminalList es) (mapDico nommer ft)
          es       = fmap (etatsND !) noms
	  ft       = fusionT es


estTerminalList :: [EtatND] -> Bool
estTerminalList es = any getFinalND es
 

regroupe :: [TransND] -> TransND
regroupe = (mapDico concat) . unionCons 

fusionT :: [EtatND] -> TransND
fusionT = regroupe . (fmap getTransND)

nommer :: [NomND] -> Nom
nommer = hash . S.toList . S.fromList



-- créer un automate qui reconnaît un caractère compris dans un segment donné
segment :: Char -> Char -> AutomateFiniND
segment c1 c2 = AutomateFiniND (M.fromList [(hashWithSalt 5555 (c1,'a',c2),etat1) , (hashWithSalt 5555(c1,'b',c2),etat2) ])
                               [hashWithSalt 5555 (c1,'a',c2)]
               where
	          etat1 = EtatND False (D.insererIntervalle D.creer c1 c2 [hashWithSalt 5555(c1,'b',c2)])
	          etat2 = EtatND True D.creer


-- créer un automate qui reconnaît la somme des langages reconnus par les automates entres
(!+!) :: AutomateFiniND -> AutomateFiniND -> AutomateFiniND
e1 !+! e2 =
     AutomateFiniND (M.union etats1 etats2) (i1 ++ i2)
     where
      (AutomateFiniND etats1 i1) = modifNom e1 (\x -> hashWithSalt (x*235154) 'a')
      (AutomateFiniND etats2 i2) = modifNom e2 (\x -> hashWithSalt (x*235154) 'b')
      
     
modifNom :: AutomateFiniND -> (Nom->Nom) -> AutomateFiniND
modifNom (AutomateFiniND etats i) f
       = AutomateFiniND (mapKeys f (fmap (modifNomE f) etats)) (fmap f i)
modifNomE :: (Nom->Nom) -> EtatND -> EtatND
modifNomE f (EtatND b tr) = EtatND b (mapDico (fmap f) tr)

-- créer un automate qui reconnaît l'étoile du langage reconnu par l'automate entre
star :: AutomateFiniND -> AutomateFiniND
star e = AutomateFiniND (M.insert (-9999999999999) etat0 (fmap cont etats)) [(-9999999999999)]
     where
       cont (EtatND False tr) = EtatND False tr
       cont (EtatND True tr) = EtatND True (regroupe [tr,transInit])
       AutomateFiniND etats i = modifNom e (\x -> hashWithSalt x 's')
       transInit = regroupe (fmap (getTransND . (etats !)) i)
       etat0 = EtatND True transInit


-- créer un automate qui reconnaît la concaténation des langages reconnus par les automates entres
(!.!) :: AutomateFiniND -> AutomateFiniND -> AutomateFiniND
e1 !.! e2 =
     AutomateFiniND (M.union (fmap cont etats1) etats2) i1
     where
       cont (EtatND False tr) = EtatND False tr
       cont (EtatND True  tr) = EtatND (estTerminalList  (fmap (etats2!) i2))
                                       (regroupe (tr:(fmap (getTransND . (etats2 !)) i2 )))
       AutomateFiniND etats1 i1 = modifNom e1 (\x -> hashWithSalt x 'a')
       AutomateFiniND etats2 i2 = modifNom e2 (\x -> hashWithSalt x 'b')

-- type des expressions régulières
data Regexp = Segment Char Char  |  Regexp :+: Regexp  |  Star Regexp  |  Regexp :.: Regexp |  Eps


-- construire un automate reconnaissant le même langage que l'expression régulière entrée
regexp2Automata :: Regexp -> AutomateFini
regexp2Automata = determinise . regexp2AutomataND

regexp2AutomataND :: Regexp -> AutomateFiniND
regexp2AutomataND (Segment c1 c2) = segment c1 c2
regexp2AutomataND (r1 :+: r2) = regexp2AutomataND r1 !+! regexp2AutomataND r2
regexp2AutomataND (r1 :.: r2) = regexp2AutomataND r1 !.! regexp2AutomataND r2
regexp2AutomataND (Star r) = star (regexp2AutomataND r)
regexp2AutomataND Eps =  AutomateFiniND (M.fromList [(999999999,EtatND True D.creer)]) [999999999]





-- Automates figés

type TransitionsF = Dictionnaire Char EtatF
data EtatF = EtatF Bool TransitionsF 
type AutomateFiniF = EtatF



figer :: AutomateFini -> AutomateFiniF
figer (AutomateFini etats i) = figerE (etats ! i)
  where
  figerE (Etat b tr) = EtatF b (mapDicoUnsafe (figerE . (etats !)) tr)

reconiseF :: AutomateFiniF -> String -> Bool
reconiseF e w = case (reconiseFM e w) of
      Nothing -> False
      Just b  -> b
  where
  reconiseFM :: AutomateFiniF -> String -> Maybe Bool
  reconiseFM (EtatF b _) []  = Just b
  reconiseFM (EtatF _ tr) (c:w) = do
    etat <- rechercherParCle  tr c
    reconiseFM etat w







-- minimiser un automate

minimisation :: AutomateFini -> AutomateFini
minimisation aut@(AutomateFini etats i) = AutomateFini etatsMin (transfNoms ! i)
  where
  transfNoms = transformation aut
  etatsMin = foldlWithKey (\acc n1 n2 -> M.insert n2 (transfEtat (etats ! n1)) acc) M.empty transfNoms
  transfEtat (Etat b tr) = Etat b (mapDico (transfNoms !) tr)


transformation :: AutomateFini -> Map Nom Nom
transformation  (AutomateFini etats i) = minAux (size init) init
  where
  init = fmap (hash . getFinal) etats
  size = S.size . S.fromList . M.elems
  contr m (Etat b tr) = hash (b, mapDico (m !) tr) 
  minAux s acc
      | s == s'   = acc
      | otherwise = minAux s' res
      where
        res = fmap (contr acc) etats  
        s'  = size res


type BigString = [String]

executAll :: AutomateFini -> String -> Map Nom Nom
executAll (AutomateFini etats _) st = mapMaybeWithKey (\n _ -> executOne n st) etats
   where
   executOne :: Nom -> String -> Maybe Nom
   executOne n ""    = Just n
   executOne n (c:w) = do
       e  <- etats !? n
       n' <- rechercherParCle (getTrans e) c
       executOne n' w



-- Bonus Phase 4



paralleleExecution :: AutomateFini -> BigString -> [Map Nom Nom]
paralleleExecution aut = parMapS (executAll aut)

parReconnise :: AutomateFini -> BigString -> Bool
parReconnise aut@(AutomateFini etats i) bs =
  final (reduce i (paralleleExecution aut bs))
  where
  reduce :: Nom -> [Map Nom Nom] -> Maybe Nom
  reduce = foldM (\n d -> d !? n)
  final :: Maybe Nom -> Bool
  final Nothing   = False
  final (Just n)  = getFinal (etats ! n)



-- a utiliser si librairie indisponible et dans la version rendu

parMapS :: (a -> Map Nom Nom) -> [a] -> [Map Nom Nom]
parMapS = fmap


-- a utiliser si librairie disponible (pas dans la version rendue
--
-- evalDico :: Strategie (Map Nom Nom)
-- evalDico = evalTraversable return
--
-- parMapS :: (a -> Map Nom Nom) -> [a] -> [Map Nom Nom]
-- parMapS = parMap evalDico